-- Author: Ashnal
-- BoostLib is a library that other scripts can use to define boosts
-- using it ensures that the new boosts will work nicely with Boost Stacking and boost chaining,
-- and it doesnt require hacks like re-using the MT or sneakertimer in most cases

local prefix = "blib_"

local cv_boostreadout = CV_RegisterVar({
    name = prefix.."boostreadout",
    defaultvalue = "Off",
    PossibleValue = CV_OnOff
})

local cv_booststack = CV_RegisterVar({
    name = prefix.."booststack",
    defaultvalue = "On",
	flags = CV_NETVAR,
    PossibleValue = CV_OnOff,
})

local cv_boostchain = CV_RegisterVar({
    name = prefix.."boostchain",
    defaultvalue = "On",
	flags = CV_NETVAR,
    PossibleValue = CV_OnOff,
})

-- This is here for Acro 0.5a compat, which looks for this cvar by name
local cv_legacyboostchain = CV_RegisterVar({
    name = "slipstream_boostchain",
    defaultvalue = "On",
	flags = nil,
    PossibleValue = CV_OnOff
})

local cv_diminish = CV_RegisterVar({
    name = prefix.."diminishingreturns",
    defaultvalue = "On",
	flags = CV_NETVAR|CV_NOSHOWHELP|CV_CHEAT,
    PossibleValue = CV_OnOff,
})

local cv_diminishamount = CV_RegisterVar({
    name = prefix.."diminishingreturnsamount",
    defaultvalue = "1.70",
	flags = CV_NETVAR|CV_FLOAT,
    PossibleValue = {MIN = FRACUNIT+FRACUNIT/4, MAX = FRACUNIT*2+FRACUNIT/3},
})

local cv_debugboost = CV_RegisterVar({
    name = prefix.."debugboost",
    defaultvalue = "0",
    flags = CV_NETVAR|CV_FLOAT,
    PossibleValue = {MIN = 0, MAX = INT32_MAX},
})

local cv_debugaccel = CV_RegisterVar({
    name = prefix.."debugaccel",
    defaultvalue = "0",
    flags = CV_NETVAR|CV_FLOAT,
    PossibleValue = {MIN = 0, MAX = INT32_MAX},
})

--fucking local constants to make everything perform better, because srb2 does a string lookup for hardcode constants, which tanks performance
local FRACUNIT = FRACUNIT
local FRACBITS = FRACBITS

--kartstuff constants
local k_sneakertimer = k_sneakertimer
local k_spinouttimer = k_spinouttimer
local k_wipeoutslow = k_wipeoutslow
local k_floorboost = k_floorboost
local k_itemamount = k_itemamount
local k_itemtype = k_itemtype
local k_rocketsneakertimer = k_rocketsneakertimer
local k_drift = k_drift
local k_speedboost = k_speedboost
local k_accelboost = k_accelboost
local k_aizdriftstrat = k_aizdriftstrat
local k_offroad = k_offroad
local k_driftboost = k_driftboost
local k_boostpower = k_boostpower
local k_hyudorotimer = k_hyudorotimer
local k_invincibilitytimer = k_invincibilitytimer

local starttime = 6*TICRATE + (3*TICRATE/4)

local speedboost = 0
local accelboost = 0

local pmo,pks,tks

freeslot("SPR_BSSS")
freeslot("S_STACKINGVFX1")
local S_STACKINGVFX1 = S_STACKINGVFX1
states[S_STACKINGVFX1] = {SPR_BSSS, FF_TRANS30|FF_ANIMATE|A, -1, nil, 4, 2, S_STACKINGVFX1}

local function SpawnStackVFX(mo, color)
	local stack = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	stack.state = S_STACKINGVFX1
	stack.flags = MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_DONTENCOREMAP|MF_NOBLOCKMAP
	stack.radius = 24*FRACUNIT
	stack.height = 100*FRACUNIT
	stack.target = mo
	stack.color = color
	stack.colorized = true
	stack.stackingvfx = true
	return stack
end

-- prefer to input all fixed values
-- https://en.wikipedia.org/wiki/Feature_scaling
local function Rescale(value, oldmin, oldmax, newmin, newmax)
	return newmin +  FixedMul(FixedDiv( value-oldmin , oldmax-oldmin), newmax-newmin)
end

local function Clamp(value, newmin ,newmax)
	return min(newmax,max(newmin,value))
end

local travelangle
local newx,newy
addHook("MobjThinker", function(mo)
	if not mo.stackingvfx then return end
	if not (mo.target and mo.target.valid and mo.target.player and mo.target.player.STACKINGVFX and not mo.player) then
		P_RemoveMobj(mo)
	else
		travelangle = mo.target.angle-(ANG1*9)*mo.target.player.kartstuff[k_drift]
		newx = mo.target.x + P_ReturnThrustX(mo.target, travelangle, FixedMul(20*FRACUNIT, mo.scale))
		newy = mo.target.y + P_ReturnThrustY(mo.target, travelangle, FixedMul(20*FRACUNIT, mo.scale))
		mo.destscale = FixedMul(mo.target.scale, Rescale(mo.target.player.vfxstackedboosts*FRACUNIT, 2*FRACUNIT, 4*FRACUNIT,  FRACUNIT, 2*FRACUNIT))
		P_MoveOrigin(mo, newx, newy, mo.target.z - (FixedMul(mo.info.height, mo.scale) - FixedMul(mo.info.height,mo.target.scale))/2)
	end
end, MT_THOK)

--lets just reimplement boosting lololololol
--by callmore#4947 and Ashnal#4043
local slopeboost
hud.add(function(v, p, c)
    if p.spectator then return end
	if not cv_boostreadout.value then return end
	if p.BLcustomboosts and p.BLcustomboosts["slope"] then slopeboost = FixedMul(p.BLcustomboosts["slope"].speedboost, 100*FRACUNIT)>>FRACBITS else slopeboost = 0 end
	v.drawString(12,25, "    velz: " .. (p.mo.momz>>FRACBITS or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,35, "    trkt: " .. (p.trickboost or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,45, "    dftt: " .. p.kartstuff[k_driftboost],V_SNAPTOLEFT,"left")
	v.drawString(12,55, "    snkt: " .. p.kartstuff[k_sneakertimer],V_SNAPTOLEFT,"left")
	v.drawString(12,65, "nisnkt: " .. (p.mo and p.mo.naturalitemsneakertimer or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,75, "  isnkt: " .. (p.mo and p.mo.itemsneakertimer or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,85, "npsnkt: " .. (p.mo and p.mo.naturalpanelsneakertimer or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,95, "  psnkt: " .. (p.mo and p.mo.panelsneakertimer or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,105, "      bst: " .. FixedMul(p.kartstuff[k_boostpower], 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
	v.drawString(12,115, "    sspd: " .. (slopeboost or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,125, "    tspd: " .. FixedMul(p.totalboost, 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
    v.drawString(12,135, "    dspd: " .. FixedMul(p.kartstuff[k_speedboost], 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
	v.drawString(12,145, "      acc: " .. FixedMul(p.kartstuff[k_accelboost], 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
end)

local BLib = {}
local psscb
local chainfuncs = {}

BLib.RegisterChainFunc = function(name, chainfunc)
	chainfuncs[name] = chainfunc
end

BLib.SetBoost = function(player, name, speedboost, accelboost, boostmult, stack, stackvfx, chain)
	if not (player and name and leveltime > 0) then return end
	if not player.BLcustomboosts then player.BLcustomboosts = {} end
	if not player.BLcustomboosts[name] then player.BLcustomboosts[name] = {} end
	psscb = player.BLcustomboosts
	psscb[name].speedboost = speedboost or 0
	psscb[name].accelboost = accelboost or 0
	psscb[name].boostmult = boostmult or 0
	psscb[name].chain = chain or false
	-- Chainfunc registration is done on load
	if stack == nil then psscb[name].stack = true else psscb[name].stack = stack end
	if stackvfx == nil then psscb[name].stackvfx = true else psscb[name].stackvfx = stackvfx end
end

local highesttimer
BLib.BoostChain = function(player, name, timer)
	if not cv_boostchain.value and player and player.BLcustomboosts then return timer end
	highesttimer = 0
	for k,v in pairs(player.BLcustomboosts or {}) do
		if k == name or not v.chain then continue end -- Don't call your own chain function, or run chain functions for stuff that doesn't chain
		highesttimer = max(chainfuncs[k](player, name, timer) or 0, $) -- Note that this calls the chainfunc for each boost -- also, let it fail if no chainfunc is defined, so any plugin authors get aan error message
	end
	return highesttimer
end

rawset(_G, "BLib", BLib)

local sneakerboost = { 53740+768, 32768, 17294+768}
local maxvanillaboost = { 53740+768, 32768, FRACUNIT*375/1000 }
local maxairspeedtarget = FRACUNIT*50
local scaledmaxairspeedtarget
local speedboost,accelboost,boostmult,speedmod,boostmod,intermediate
local nonstackboost, nonstackboostmult, hasChainableBoost, harddiminish, onground
local function GetKartBoostPower(player)

	if pks[k_spinouttimer] and pks[k_wipeoutslow] == 1 then -- Slow down after you've been bumped
		pks[k_speedboost] = 0
		pks[k_accelboost] = 0
		return
	end

	speedboost = 0
	nonstackboost = 0
	accelboost = 0
	boostmult = FRACUNIT
	nonstackboostmult = FRACUNIT
	hasChainableBoost = false
	player.vfxstackedboosts = 0
	player.stackedboosts = 0

	if cv_debugboost.value then
		speedboost = $ + cv_debugboost.value
	end
	if cv_debugaccel.value then
		accelboost = max($, cv_debugaccel.value)
	end
	
	--print("- boosttable - "..player.name)
	if player.BLcustomboosts then
		for k,v in pairs(player.BLcustomboosts) do			
			
			speedmod = v.speedboost -- Can't be nil
			boostmod = v.boostmult
		
			--print(k.." "..speedmod.." "..boostmod.." "..v.accelboost)
		
			if (speedmod or boostmod) and cv_boostchain.value and v.chain then 
				hasChainableBoost = true 
			end
		
			if (speedmod or boostmod) and cv_booststack.value then
				player.stackedboosts = $ + 1
				if v.stackvfx then
					player.vfxstackedboosts = $ + 1
				end
			end
			if v.stack and cv_booststack.value then
				speedboost = $ + speedmod
				boostmult = $ + boostmod
			else
				nonstackboost = max($, speedmod)
				nonstackboostmult = max($, FRACUNIT + boostmod)
			end
				
			accelboost = max($, v.accelboost)
		end
		
		speedboost = max($, nonstackboost)
		
		if nonstackboostmult ~= FRACUNIT then -- we had a nonstacking boostmult mod -- this wierdness allows for negative stacking boostmult as long as no other nonstacking boostmult mods are active
			boostmult = max($, nonstackboostmult)
		end
	end

	onground = P_IsObjectOnGround(pmo)

	if  player.vfxstackedboosts >= 2 then
		--SpawnFastLines(player, 5*FRACUNIT/3, player.skincolor+12)		
		if not player.STACKINGVFX and pmo then
			player.STACKINGVFX = SpawnStackVFX(pmo, player.skincolor)
		end
		pmo.dotrail = true
	else
		player.STACKINGVFX = nil
		pmo.dotrail = false
	end
	if not onground then 
		pmo.dotrail = false 
	end
	
	
	player.totalboost = speedboost -- saved for later display by boostreadout

	-- WARNING: Math ahead
	-- Apply new diminishing returns equation -- OLD: https://www.desmos.com/calculator/eoxcakxvxy
	-- NEW: https://www.desmos.com/calculator/wnc8uw72m9
	if cv_diminish.value then  -- two stage, linear unti 0.5, then we start falling off
		if gamespeed <= 1 and speedboost > FRACUNIT/2 then -- Normal, green line
			intermediate = FixedDiv(FixedMul(cv_diminishamount.value,FRACUNIT*-1/2) - FRACUNIT/4,-cv_diminishamount.value+FRACUNIT/2)
			speedboost = FixedMul(cv_diminishamount.value,(FRACUNIT-FixedDiv(FRACUNIT,($+intermediate))))
		elseif gamespeed == 2 and speedboost > FRACUNIT*375/1000 then -- We hard speed, use the harsher curve, red line
			harddiminish = Rescale(cv_diminishamount.value, FRACUNIT, 2*FRACUNIT, FRACUNIT*95/100, FRACUNIT*180/100)
			intermediate = FixedDiv(FixedMul(harddiminish,FRACUNIT*-625/1000) - 9216,-harddiminish+FRACUNIT*375/1000)
			speedboost = FixedMul(harddiminish,(FRACUNIT-FixedDiv(FRACUNIT,($+intermediate))))
		end
	end

	if pks[k_offroad] and not pks[k_hyudorotimer] and not pks[k_invincibilitytimer] and not pmo.naturalpanelsneakertimer and not pmo.naturalitemsneakertimer then -- mow nerf. We behave as if the player has no boosts while in offroad, and smooth down using the vanila rate
		pks[k_speedboost] = max(player.SSlastframespeedboost - player.SSlastframespeedboost/12, FRACUNIT*18/100) -- Apply strong falloff when in offroad and limit boost to a bit less than MT
		pks[k_accelboost] = accelboost -- You get to keep your accel
	elseif player.cmd.buttons & BT_BRAKE then
		pks[k_speedboost] = max(player.SSlastframespeedboost - (FRACUNIT/90 + player.SSlastframespeedboost/25), min($, maxvanillaboost[gamespeed+1]))
		pks[k_accelboost] = accelboost
	elseif pks[k_aizdriftstrat] and abs(pks[k_drift]) < 5 then
		pks[k_speedboost] = max(player.SSlastframespeedboost - (FRACUNIT/90 + player.SSlastframespeedboost/25), min($, maxvanillaboost[gamespeed+1]))
		pks[k_accelboost] = accelboost
	elseif speedboost >= (player.SSlastframespeedboost or 0) then
		pks[k_speedboost] = speedboost
		pks[k_accelboost] = accelboost
	else
		pks[k_speedboost] = max(player.SSlastframespeedboost - FRACUNIT/90, 0) -- value smoothing
		pks[k_accelboost] = accelboost
	end
	
	-- This here is the boostmult, its implemented as an adjustment to boostpower
	pks[k_boostpower] = pks[k_boostpower] + (FixedMul(pks[k_speedboost], boostmult) - pks[k_speedboost])
	
	--player.speed = min($, FixedMul(FRACUNIT*128, mapobjectscale)) -- Turn right bug starts occuring somewhere between 130 and 140 FU/T, so we hard limit speed to prevent that
	--if pks[k_pogospring] then player.speed = min($, FixedMul(FRACUNIT*75, mapobjectscale)) end

	if not onground and pmo.lastonground then -- We launched
		pmo.launchspeed = player.speed
		pmo.lastairspeed = player.speed
		pmo.launchspeedz = pmo.momz
		pmo.timeoffground = 0
	end
	if not onground then
		pmo.timeoffground = ($ or 0) + 1
		pmo.lastairspeed = ($ or 0)
		player.speed = min(max(pmo.lastairspeed - (pmo.lastairspeed - scaledmaxairspeedtarget)/70, scaledmaxairspeedtarget), pmo.lastairspeed)
		pmo.lastairspeed = player.speed
	else
		pmo.timeoffground = 0
	end
	pmo.lastonground = onground
	
	-- Award stuff
	player.highestboost = max($ or 0, pks[k_speedboost])
	player.higheststackcount = max($ or 0, player.stackedboosts)
	if hasChainableBoost then
		player.currentchainduration = ($ or 0) + 1
		player.longestchain = max($ or 0, player.currentchainduration)
	else
		player.currentchainduration = 0
	end
	
	player.SSlastframespeedboost = pks[k_speedboost]
end

addHook("MapChange", function(mapnum)
	for player in players.iterate do
		player.longestchain = 0
		player.higheststackcount = 0
		player.BLcustomboosts = {}
	end
end)

-- HACKS! Hopefully allows someone to retain their turning radius while schmovin
-- reduces speed while sliptiding to reasonable levels to retain control
/*
local sneakerboost = { 53740+768, 32768, 17294+768}
local sneakerspeed, sliptidespeed, adjustedspeed
addHook("MobjThinker", function(mo)
	if mo.player.kartstuff[k_aizdriftstrat] and not mo.player.kartstuff[k_drift] then
		sneakerspeed = FixedMul(K_GetKartSpeed(mo.player, false), FRACUNIT+sneakerboost[gamespeed+1])
		sliptidespeed = sneakerspeed*86/100-(mo.player.kartspeed<<FRACBITS) -- this comes out to right about 50 FU/T for most characters. I've observed all the vanilla char sliptiding between 49 and 51 on normal. SLiptides end up slightly slower on Hard for some reason
		adjustedspeed = 0 --sliptidespeed*40/100 -- account for acceleration
		-- only reduce spoeed if it is more than sliptidespeed
		-- reduce it by 1/12 per tic if it is above sliptidespeed
		mo.player.speed = min($,max(FixedMul($,FRACUNIT-FRACUNIT/10),adjustedspeed))
		--print("KartSpeed: "..K_GetKartSpeed(mo.player, false)>>FRACBITS.." sneakerspeed: "..sneakerspeed>>FRACBITS.." sliptidespeed: "..sliptidespeed>>FRACBITS.." actual speed: "..mo.player.speed>>FRACBITS)
	end
end, MT_PLAYER)
*/

addHook("PlayerSpawn", function(p)
	p.BLcustomboosts = {}
end)

addHook("ThinkFrame", do
	-- stupid hack for Acro's detection
	if  cv_legacyboostchain.value ~= cv_boostchain.value then
		COM_BufInsertText(consoleplayer, "slipstream_boostchain "..cv_boostchain.value)
	end

	scaledmaxairspeedtarget = FixedMul(maxairspeedtarget, mapobjectscale)
	for p in players.iterate do	
		pmo = p.mo
		pks = p.kartstuff
		if pmo then
			GetKartBoostPower(p)
		end
	end
end)

--hostmod scoreboard
local function hmaddmod(fug)
	table.insert(server.HMNSBbalance, fug)
end

local hmload = false
local function hostmodload()
	if not hmload and HOSTMOD then
		hmaddmod({disp = "Boost Stacking", var = prefix.."booststack"})
		hmaddmod({disp = "Boost Chaining", var = prefix.."boostchain"})
		hmload  = true
	end
end

addHook("MapLoad", hostmodload)
addHook("NetVars", hostmodload) -- Joiners don't run MapLoad