-- Author: Ashnal
-- Adds slipstreaming aka drafting like the later Mario Kart games
-- Youll see small gray speed lines when you're charging a slipstream boost
-- Theres a subtle woosh sound when it activates, and the gray speed lines will become normal size and more frequent
-- You need to have another racer within the maxangle of your heading, and be within the maxdistiance range to charge it
-- the closer you are to them, the faster it will charge
-- KNOWN BUG: the fast lines have some wierd angles on OpenGL rendering. No idea why.

local prefix = "ss_"
local cv_slipstream = CV_RegisterVar({
    name = prefix.."enabled",
    defaultvalue = "On",
    flags = CV_NETVAR,
    PossibleValue = CV_OnOff
})

local cv_colorized = CV_RegisterVar({
    name = prefix.."colorized",
    defaultvalue = "On",
    flags = NULL,
    PossibleValue = CV_OnOff
})

local cv_reminder = CV_RegisterVar({
    name = prefix.."reminder",
    defaultvalue = "On",
    flags = CV_NETVAR,
    PossibleValue = CV_OnOff
})

local cv_maxdistance = CV_RegisterVar({
    name = prefix.."maxdistance",
    defaultvalue = 2400, -- enough room to charge slowly, and dodge
    flags = CV_NETVAR,
    PossibleValue = CV_Unsigned
})

local cv_t2chargedist = CV_RegisterVar({
    name = prefix.."t2chargedist",
    defaultvalue = 1400,
    flags = CV_NETVAR,
    PossibleValue = CV_Unsigned
})

local cv_t3chargedist = CV_RegisterVar({
    name = prefix.."t3chargedist",
    defaultvalue = 300, -- danger close, woudln't be able to dodge an aimed item
    flags = CV_NETVAR,
    PossibleValue = CV_Unsigned
})

local cv_maxangle = CV_RegisterVar({
    name = prefix.."maxangle",
    defaultvalue = "10", --10 degrees
    flags = CV_NETVAR|CV_FLOAT,
    PossibleValue = CV_Unsigned
})

-- This is measured in tics
local cv_chargetoboost = CV_RegisterVar({
    name = prefix.."chargetoboost",
    defaultvalue = 5*TICRATE/2,
    flags = CV_NETVAR,
    PossibleValue = CV_Unsigned
})

local cv_boosttime = CV_RegisterVar({
    name = prefix.."boosttimer",
    defaultvalue = 2*TICRATE,
    flags = CV_NETVAR,
    PossibleValue = CV_Unsigned
})

local cv_minimumspeed = CV_RegisterVar({
    name = prefix.."minimumspeed",
    defaultvalue = 28,
    flags = CV_NETVAR,
    PossibleValue = CV_Unsigned
})

-- Boost youll get if you're a 1 kartspeed
local cv_maxspeedboostpercent = CV_RegisterVar({
    name = prefix.."maxspeedboostpercent",
    defaultvalue = "0.27",
    flags = CV_NETVAR|CV_FLOAT,
    PossibleValue = {MIN = 0, MAX = INT32_MAX}
})

-- Boost youll get if you're a 9 or higher kartspeed
local cv_minspeedboostpercent = CV_RegisterVar({
    name = prefix.."minspeedboostpercent",
    defaultvalue = "0.17",
    flags = CV_NETVAR|CV_FLOAT,
    PossibleValue = {MIN = 0, MAX = INT32_MAX}
})

local cv_accelboostpercent = CV_RegisterVar({
    name = prefix.."accelboostpercent",
    defaultvalue = "0.50",
    flags = CV_NETVAR|CV_FLOAT,
    PossibleValue = {MIN = 0, MAX = INT32_MAX}
})

local cv_maxdraftspeedboostpercent = CV_RegisterVar({
    name = prefix.."maxdraftspeedboostpercent",
    defaultvalue = "0.12",
    flags = CV_NETVAR|CV_FLOAT,
    PossibleValue = {MIN = 0, MAX = INT32_MAX}
})

local cv_mindraftspeedboostpercent = CV_RegisterVar({
    name = prefix.."mindraftspeedboostpercent",
    defaultvalue = "0.01",
    flags = CV_NETVAR|CV_FLOAT,
    PossibleValue = {MIN = 0, MAX = INT32_MAX}
})

local cv_anglefinder = CV_RegisterVar({
    name = prefix.."anglefinder",
    defaultvalue = "Off",
    PossibleValue = CV_OnOff
})

local cv_alwayson = CV_RegisterVar({
    name = prefix.."alwaysactive",
    defaultvalue = "Off",
	flags = CV_NETVAR,
    PossibleValue = CV_OnOff
})

freeslot(
"MT_ANGLEFINDER",
"S_ANGLEFINDER"
)

states[S_ANGLEFINDER] = {SPR_THOK, FF_TRANS30, 8, nil, 0, 0, S_ANGLEFINDER}

mobjinfo[MT_ANGLEFINDER] = {
	spawnstate = S_ANGLEFINDER,
	flags = MF_NOGRAVITY|MF_DONTENCOREMAP|MF_NOBLOCKMAP,
	radius = 1*FRACUNIT,
	height = 1*FRACUNIT
}

local INT32_MAX = INT32_MAX
local TICRATE = TICRATE
local FRACUNIT = FRACUNIT
local ANGLE_135 = ANGLE_135
local ANGLE_45 = ANGLE_45
local FRACBITS = FRACBITS
local SKINCOLOR_WHITE = SKINCOLOR_WHITE
local MT_FASTLINE = MT_FASTLINE
local MT_PARTICLE = MT_PARTICLE

local k_hyudorotimer = k_hyudorotimer
local k_spinouttimer = k_spinouttimer
local k_wipeoutslow = k_wipeoutslow
local k_squishedtimer = k_squishedtimer
local k_drift = k_drift
local k_offroad = k_offroad

-- prefer to input all fixed values
-- https://en.wikipedia.org/wiki/Feature_scaling
local function Rescale(value, oldmin, oldmax, newmin, newmax)
	return newmin +  FixedMul(FixedDiv( value-oldmin , oldmax-oldmin), newmax-newmin)
end

local function Clamp(value, newmin ,newmax)
	return min(newmax,max(newmin,value))
end

local newangleoffset = 0
local newx, newy, seekangle, tempdist, driftmod
addHook("MobjThinker", function(mo)
	if not mo.target or not cv_anglefinder.value then
		P_RemoveMobj(mo)
	else
		seekangle = mo.target.player.mo.angle-(131235115)*mo.target.player.kartstuff[k_drift]
		
		if mo.target.player.kartstuff[k_drift] then 
			tempdist = FixedMul(mo.distanceoffset, 2*FRACUNIT/5)
			newangleoffset = mo.angleoffset * FixedAngle(cv_maxangle.value*2)
		else 
			tempdist = FixedMul(mo.distanceoffset, FRACUNIT)
			newangleoffset = mo.angleoffset * FixedAngle(cv_maxangle.value)
		end

		newx = mo.target.x + P_ReturnThrustX(mo.target, seekangle+newangleoffset, FixedMul(tempdist, mo.target.scale))
		newy = mo.target.y + P_ReturnThrustY(mo.target, seekangle+newangleoffset, FixedMul(tempdist, mo.target.scale))
		P_TeleportMove(mo, newx, newy, mo.target.z)
	end
end, MT_ANGLEFINDER)

local closeleft,closeright,closecenter
local farleft, farright, farcenter
local function AngleFinder()
	if cv_anglefinder.value and displayplayers[0].mo then
		if not displayplayers[0].anglefinder then
		
			closeleft = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			closecenter = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			closeright = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			farleft = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			farcenter = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			farright = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			
			displayplayers[0].anglefinder = true
		end
		closeleft.target = displayplayers[0].mo
		closecenter.target = displayplayers[0].mo
		closeright.target = displayplayers[0].mo
		farleft.target = displayplayers[0].mo
		farcenter.target = displayplayers[0].mo
		farright.target = displayplayers[0].mo
		closeleft.scale = FRACUNIT/4
		closecenter.scale = FRACUNIT/4
		closeright.scale = FRACUNIT/4
		farleft.scale = FRACUNIT/4
		farcenter.scale = FRACUNIT/4
		farright.scale = FRACUNIT/4
		closeleft.distanceoffset = cv_t3chargedist.value*FRACUNIT
		closecenter.distanceoffset = cv_t3chargedist.value*FRACUNIT
		closeright.distanceoffset = cv_t3chargedist.value*FRACUNIT
		farleft.distanceoffset = cv_maxdistance.value*FRACUNIT
		farcenter.distanceoffset = cv_maxdistance.value*FRACUNIT
		farright.distanceoffset = cv_maxdistance.value*FRACUNIT
		closeleft.angleoffset = -1
		closecenter.angleoffset = 0
		closeright.angleoffset = 1
		farleft.angleoffset = -1
		farcenter.angleoffset = 0
		farright.angleoffset = 1
		closeleft.travel = true
		closecenter.travel = true
		closeright.travel = true
		farleft.travel = true
		farcenter.travel = true
		farright.travel = true
	else
		if displayplayers[0] then
			displayplayers[0].anglefinder = false
		end
	end
end

//bless Lumyni
local P_PseudoRandom = function(min,max,speed,time)
    if not min then min = 1 end
    if not max then max = 2 end
    if not speed then speed = 5000 end
    if not time then time = leveltime end

    speed = $*FRACUNIT
    local average = ((min+max) or 1)/2
    local intensity = (FRACUNIT-average)/(average or 1)
    local formula = (sin(time*speed)/intensity)+min+average
    if formula and (time%8 == 0) then formula = $-1 end
    
    return formula
end

local fast
local function SpawnFastLines(mo, scale, color)
	fast = P_SpawnMobj((mo.x - 37*mo.scale) + (P_PseudoRandom(1,72,15000) * mo.scale),
					   (mo.y - 37*mo.scale) + (P_PseudoRandom(1,72,20000) * mo.scale),
					   (mo.z - 21*mo.scale) + (mo.height/2) + (P_PseudoRandom(1,40,37000) * mo.scale),
					   MT_FASTLINE)
	fast.angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	fast.momx = 3*mo.momx/4
	fast.momy = 3*mo.momy/4
	fast.momz = 3*mo.momz/4
	fast.color = color
	fast.colorized = true
	fast.scale = FixedMul(scale,mo.scale)
	K_MatchGenericExtraFlags(fast, mo)
end

local angle,parts,i,verticalflip
local function SSDust(mo)
	angle = mo.angle
	verticalflip = mo.eflags & MFE_VERTICALFLIP
	if leveltime%2 then
		for i = 0,1 do
			newx = mo.x + P_ReturnThrustX(mo, angle + ((i == 0) and -1 or 1)*ANGLE_135, FixedMul(22*FRACUNIT, mo.scale))			
			newy = mo.y + P_ReturnThrustY(mo, angle + ((i == 0) and -1 or 1)*ANGLE_135, FixedMul(22*FRACUNIT, mo.scale))
			parts = P_SpawnMobj(newx, newy, (verticalflip and (mo.ceilingz+6*FRACUNIT*mapobjectscale or -1) or mo.floorz-6*FRACUNIT*mapobjectscale), MT_PARTICLE)
			parts.target = mo
			parts.angle = angle - ((i == 0) and -1 or 1)*ANGLE_45
			parts.scale = mo.scale
			parts.destscale = mo.scale*3
			parts.scalespeed = mo.scale/6
			parts.momx = 3*mo.momx/5
			parts.momy = 3*mo.momy/5				
			parts.fuse = 5
		end
	end
end


local T3_DISTANCE,T2_DISTANCE,MAX_DISTANCE
local clampedkartspeed,skincolor,draftingboost,slipstreamboost,slipstreamaccel,slipstreamtarget,angletotarget,disttotarget,seekangle,charge
local p,pks,tmo
addHook("MobjThinker", function(mo)
	if not (cv_slipstream.value and BLib) then	return end -- Has to be turned on

	p = mo.player
	pks = mo.player.kartstuff

	if leveltime == 3*TICRATE and cv_reminder.value and not p.slipstreamreminded then 
		chatprintf(p, "\131* You can \130draft\131 behind another player to charge a \130slipstream speed boost!", true)
		p.slipstreamreminded = true
	end
	
	if mo.state == S_KART_SPIN or pks[k_squishedtimer] then -- no slipstreaming if you've bumped or spun out
		mo.slipstreamtimer = 0
		mo.slipstreamcharge = 0
		mo.slipstreamsoundtimer = false
		BLib.SetBoost(p, "slipstream")
		return
	end

	if mo.slipstreamcharge == nil then
		mo.slipstreamcharge = 0
		mo.slipstreamsoundtimer = false
	end

    T3_DISTANCE = cv_t3chargedist.value * mapobjectscale
    T2_DISTANCE = cv_t2chargedist.value * mapobjectscale
    MAX_DISTANCE = cv_maxdistance.value * mapobjectscale

	-- reset until we find a valid slipstreamtarget this frame
	slipstreamboost = 0
	slipstreamaccel = 0
	slipstreamtarget = nil

	if (P_IsObjectOnGround(mo)
	and FixedDiv(p.speed, mapobjectscale)/FRACUNIT >= cv_minimumspeed.value -- must be moving decently on the ground
	) then
		
		for potentialtarget in players.iterate do          
			tmo = potentialtarget.mo
			
			if (tmo and tmo.valid -- must have valid player mapobject
			and tmo ~= mo -- Can't splipstream off yourself
			and not potentialtarget.kartstuff[k_hyudorotimer] -- or ghosts
			and P_IsObjectOnGround(tmo) -- or airbourne karts
			and FixedDiv(potentialtarget.speed, mapobjectscale)/FRACUNIT >= cv_minimumspeed.value -- or slowpokes
			) then
				
				-- Calculating the angle at which we look for suitable draft targets based on drifting
				seekangle = mo.angle-(131235115)*pks[k_drift] -- The magic number is 11 degrees, I'm optimizing this way >_>
				angletotarget = abs(seekangle - R_PointToAngle2(mo.x, mo.y, tmo.x, tmo.y))
				disttotarget = FixedHypot(mo.x - tmo.x, mo.y - tmo.y)
				
				if pks[k_drift] then
					if angletotarget > FixedAngle(cv_maxangle.value*2) then continue end
					if disttotarget > 2*MAX_DISTANCE/5 then continue end -- max slipstream distance while drifting
				else 
					if angletotarget > FixedAngle(cv_maxangle.value) then continue end
					if disttotarget > MAX_DISTANCE then continue end -- max slipstream distance
				end -- Narrow angle for following to slipstream

				--  print(p.mo.skin + " " + potentialtarget.mo.skin + " angletotarget: " + angletotarget + " disttotarget: " + disttotarget + " slipstreamtarget: " + p.slipstreamtarget)

				slipstreamtarget = potentialtarget
				break

			end
		end
	end

	-- add/remove slipstream charge
	if slipstreamtarget then
		if disttotarget > T2_DISTANCE then
			charge = 1
		elseif disttotarget > T3_DISTANCE then
			charge = 2
		else
			charge = 2
		end

		mo.slipstreamcharge = min($+charge, cv_chargetoboost.value)

		if mo.slipstreamcharge >= cv_chargetoboost.value then
			mo.slipstreamtimer = cv_boosttime.value
			if mo.slipstreamsoundtimer == false then
				S_StartSoundAtVolume(mo, sfx_s3k82, INT32_MAX)
				mo.slipstreamsoundtimer = true
			end
		end

		-- also spawn mini lines if charging to teach/show player the charging area
		if mo.slipstreamtimer == 0 then
			SpawnFastLines(mo, FRACUNIT/2, SKINCOLOR_WHITE)

			clampedkartspeed = Clamp(p.kartspeed, 1, 9)
			
			-- Set up drafting boost
			slipstreamboost = Rescale(clampedkartspeed*FRACUNIT, 9*FRACUNIT, FRACUNIT, cv_mindraftspeedboostpercent.value, cv_maxdraftspeedboostpercent.value)
			slipstreamaccel = 0
			
			mo.slipstreamsoundtimer = false
		end
	else
		mo.slipstreamcharge = max($-1, 0)
	end

	if mo.slipstreamtimer or cv_alwayson.value then
		clampedkartspeed = Clamp(p.kartspeed, 1, 9)
		slipstreamboost = Rescale(clampedkartspeed*FRACUNIT, 9*FRACUNIT, FRACUNIT, cv_minspeedboostpercent.value, cv_maxspeedboostpercent.value)
		slipstreamaccel = cv_accelboostpercent.value

		if cv_colorized.value and p.skincolor then
			skincolor = p.skincolor
		else 
			skincolor = SKINCOLOR_WHITE
		end
		
		SpawnFastLines(mo, FRACUNIT, skincolor)
		
		if P_IsObjectOnGround(mo) then
			SSDust(mo)
		end
	end

	slipstreamboost = min((mo.lastssboost or 0) + $/25, $) -- smoothing, takes some tics to reach full effect
	BLib.SetBoost(p, "slipstream", slipstreamboost, slipstreamaccel, nil, true, true, false)
	mo.slipstreamtimer = max(($ or 0)-1, 0)
	mo.lastssboost = slipstreamboost
end, MT_PLAYER)

addHook("ThinkFrame", do
	AngleFinder()
end)

--hostmod scoreboard
local function hmaddmod(fug)
	table.insert(server.HMNSBbalance, fug)
end

local hmload = false
local function hostmodload()
	if not hmload and HOSTMOD then
		hmaddmod({disp = "Slipstream", var = prefix.."enabled"})
		hmload  = true
	end
end

addHook("MapLoad", hostmodload)
addHook("NetVars", hostmodload) -- Joiners don't run MapLoad